/*
 * tests.h
 *
 *  Created on: 22.01.2017
 *      Author: hoene
 */

#ifndef TESTS_H_
#define TESTS_H_

#include <CUnit/Basic.h>
#include <CUnit/CUnit.h>

void test_coordinates();

void test_nsearch();

void test_check();

void test_check_data();

void test_lookup();

void test_neighbors();

void test_interpolate();

void test_resample();

void test_loudness();

void test_minphase();

void test_cache();

void test_easy();

void test_easy_open();

void test_easy_nonorm();

void test_user_defined_variable();

#endif /* TESTS_H_ */
